import os
from pathlib import Path
from typing import Type

import onnxruntime as ort

from .session_base import BaseSession

# from .session_cloth import ClothSession
from .session_simple import SimpleSession


def new_session(model_name: str = "u2net") -> BaseSession:
    session_class: Type[BaseSession]
    # md5 = "60024c5c889badc19c04ad937298a77b"
    # url = "https://github.com/danielgatis/rembg/releases/download/v0.0.0/u2net.onnx"
    session_class = SimpleSession

    if model_name != "u2net":
        raise NotImplementedError

    u2net_home = os.getenv("U2NET_HOME") or Path(__file__).parent

    fname = f"{model_name}.onnx"
    path = Path(u2net_home).expanduser()
    full_path = path / fname

    sess_opts = ort.SessionOptions()

    if "OMP_NUM_THREADS" in os.environ:
        sess_opts.inter_op_num_threads = int(os.environ["OMP_NUM_THREADS"])

    return session_class(
        model_name,
        ort.InferenceSession(
            str(full_path),
            providers=ort.get_available_providers(),
            sess_options=sess_opts,
        ),
    )



import io
from enum import Enum
from typing import List, Optional, Union

import numpy as np
from cv2 import (
    BORDER_DEFAULT,
    MORPH_ELLIPSE,
    MORPH_OPEN,
    GaussianBlur,
    getStructuringElement,
    morphologyEx,
)
from PIL import Image
from PIL.Image import Image as PILImage

# from pymatting.alpha.estimate_alpha_cf import estimate_alpha_cf # isort:skip
# from pymatting.foreground.estimate_foreground_ml import estimate_foreground_ml # isort:skip
# from pymatting.util.util import stack_images # isort:skip
# from scipy.ndimage import binary_erosion # isort:skip
from .session_base import BaseSession
from .session_factory import new_session

kernel = getStructuringElement(MORPH_ELLIPSE, (3, 3))


class ReturnType(Enum):
    BYTES = 0
    PILLOW = 1
    NDARRAY = 2


# def alpha_matting_cutout(
#     img: PILImage,
#     mask: PILImage,
#     foreground_threshold: int,
#     background_threshold: int,
#     erode_structure_size: int,
# ) -> PILImage:
#
#     if img.mode == "RGBA" or img.mode == "CMYK":
#         img = img.convert("RGB")
#
#     img = np.asarray(img)
#     mask = np.asarray(mask)
#
#     is_foreground = mask > foreground_threshold
#     is_background = mask < background_threshold
#
#     structure = None
#     if erode_structure_size > 0:
#         structure = np.ones(
#             (erode_structure_size, erode_structure_size), dtype=np.uint8
#         )
#
#     is_foreground = binary_erosion(is_foreground, structure=structure)
#     is_background = binary_erosion(is_background, structure=structure, border_value=1)
#
#     trimap = np.full(mask.shape, dtype=np.uint8, fill_value=128)
#     trimap[is_foreground] = 255
#     trimap[is_background] = 0
#
#     img_normalized = img / 255.0
#     trimap_normalized = trimap / 255.0
#
#     alpha = estimate_alpha_cf(img_normalized, trimap_normalized)
#     foreground = estimate_foreground_ml(img_normalized, alpha)
#     cutout = stack_images(foreground, alpha)
#
#     cutout = np.clip(cutout * 255, 0, 255).astype(np.uint8)
#     cutout = Image.fromarray(cutout)
#
#     return cutout


def naive_cutout(img: PILImage, mask: PILImage) -> PILImage:
    empty = Image.new("RGBA", (img.size), 0)
    cutout = Image.composite(img, empty, mask)
    return cutout


def get_concat_v_multi(imgs: List[PILImage]) -> PILImage:
    pivot = imgs.pop(0)
    for im in imgs:
        pivot = get_concat_v(pivot, im)
    return pivot


def get_concat_v(img1: PILImage, img2: PILImage) -> PILImage:
    dst = Image.new("RGBA", (img1.width, img1.height + img2.height))
    dst.paste(img1, (0, 0))
    dst.paste(img2, (0, img1.height))
    return dst


def post_process(mask: np.ndarray) -> np.ndarray:
    """
    Post Process the mask for a smooth boundary by applying Morphological Operations
    Research based on paper: https://www.sciencedirect.com/science/article/pii/S2352914821000757
    args:
        mask: Binary Numpy Mask
    """
    mask = morphologyEx(mask, MORPH_OPEN, kernel)
    mask = GaussianBlur(mask, (5, 5), sigmaX=2, sigmaY=2, borderType=BORDER_DEFAULT)
    mask = np.where(mask < 127, 0, 255).astype(np.uint8)  # convert again to binary
    return mask


def remove(
    data: Union[bytes, PILImage, np.ndarray],
    alpha_matting: bool = False,
    # alpha_matting_foreground_threshold: int = 240,
    # alpha_matting_background_threshold: int = 10,
    # alpha_matting_erode_size: int = 10,
    session: Optional[BaseSession] = None,
    only_mask: bool = False,
    post_process_mask: bool = False,
) -> Union[bytes, PILImage, np.ndarray]:
    if isinstance(data, PILImage):
        return_type = ReturnType.PILLOW
        img = data
    elif isinstance(data, bytes):
        return_type = ReturnType.BYTES
        img = Image.open(io.BytesIO(data))
    elif isinstance(data, np.ndarray):
        return_type = ReturnType.NDARRAY
        img = Image.fromarray(data)
    else:
        raise ValueError("Input type {} is not supported.".format(type(data)))

    if session is None:
        session = new_session("u2net")

    masks = session.predict(img)
    cutouts = []

    for mask in masks:
        if post_process_mask:
            mask = Image.fromarray(post_process(np.array(mask)))

        if only_mask:
            cutout = mask

        elif alpha_matting:
            raise NotImplementedError
            # try:
            #     cutout = alpha_matting_cutout(
            #         img,
            #         mask,
            #         alpha_matting_foreground_threshold,
            #         alpha_matting_background_threshold,
            #         alpha_matting_erode_size,
            #     )
            # except ValueError:
            #     cutout = naive_cutout(img, mask)

        else:
            cutout = naive_cutout(img, mask)

        cutouts.append(cutout)

    cutout = img
    if len(cutouts) > 0:
        cutout = get_concat_v_multi(cutouts)

    if ReturnType.PILLOW == return_type:
        return cutout

    if ReturnType.NDARRAY == return_type:
        return np.asarray(cutout)

    bio = io.BytesIO()
    cutout.save(bio, "PNG")
    bio.seek(0)

    return bio.read()



# import pathlib
# import sys
# import time
# from enum import Enum
# from typing import IO, cast
#
# import aiohttp
# import click
# import filetype
# import uvicorn
# from asyncer import asyncify
# from fastapi import Depends, FastAPI, File, Form, Query
# from fastapi.middleware.cors import CORSMiddleware
# from starlette.responses import Response
# from tqdm import tqdm
# from watchdog.events import FileSystemEvent, FileSystemEventHandler
# from watchdog.observers import Observer
#
# from . import _version
# from .bg import remove
# from .session_base import BaseSession
# from .session_factory import new_session
#
#
# @click.group()
# @click.version_option(version=_version.get_versions()["version"])
# def main() -> None:
#     pass
#
#
# @main.command(help="for a file as input")
# @click.option(
#     "-m",
#     "--model",
#     default="u2net",
#     type=click.Choice(
#         ["u2net", "u2netp", "u2net_human_seg", "u2net_cloth_seg", "silueta"]
#     ),
#     show_default=True,
#     show_choices=True,
#     help="model name",
# )
# @click.option(
#     "-a",
#     "--alpha-matting",
#     is_flag=True,
#     show_default=True,
#     help="use alpha matting",
# )
# @click.option(
#     "-af",
#     "--alpha-matting-foreground-threshold",
#     default=240,
#     type=int,
#     show_default=True,
#     help="trimap fg threshold",
# )
# @click.option(
#     "-ab",
#     "--alpha-matting-background-threshold",
#     default=10,
#     type=int,
#     show_default=True,
#     help="trimap bg threshold",
# )
# @click.option(
#     "-ae",
#     "--alpha-matting-erode-size",
#     default=10,
#     type=int,
#     show_default=True,
#     help="erode size",
# )
# @click.option(
#     "-om",
#     "--only-mask",
#     is_flag=True,
#     show_default=True,
#     help="output only the mask",
# )
# @click.option(
#     "-ppm",
#     "--post-process-mask",
#     is_flag=True,
#     show_default=True,
#     help="post process the mask",
# )
# @click.argument(
#     "input", default=(None if sys.stdin.isatty() else "-"), type=click.File("rb")
# )
# @click.argument(
#     "output",
#     default=(None if sys.stdin.isatty() else "-"),
#     type=click.File("wb", lazy=True),
# )
# def i(model: str, input: IO, output: IO, **kwargs) -> None:
#     output.write(remove(input.read(), session=new_session(model), **kwargs))
#
#
# @main.command(help="for a folder as input")
# @click.option(
#     "-m",
#     "--model",
#     default="u2net",
#     type=click.Choice(
#         ["u2net", "u2netp", "u2net_human_seg", "u2net_cloth_seg", "silueta"]
#     ),
#     show_default=True,
#     show_choices=True,
#     help="model name",
# )
# @click.option(
#     "-a",
#     "--alpha-matting",
#     is_flag=True,
#     show_default=True,
#     help="use alpha matting",
# )
# @click.option(
#     "-af",
#     "--alpha-matting-foreground-threshold",
#     default=240,
#     type=int,
#     show_default=True,
#     help="trimap fg threshold",
# )
# @click.option(
#     "-ab",
#     "--alpha-matting-background-threshold",
#     default=10,
#     type=int,
#     show_default=True,
#     help="trimap bg threshold",
# )
# @click.option(
#     "-ae",
#     "--alpha-matting-erode-size",
#     default=10,
#     type=int,
#     show_default=True,
#     help="erode size",
# )
# @click.option(
#     "-om",
#     "--only-mask",
#     is_flag=True,
#     show_default=True,
#     help="output only the mask",
# )
# @click.option(
#     "-ppm",
#     "--post-process-mask",
#     is_flag=True,
#     show_default=True,
#     help="post process the mask",
# )
# @click.option(
#     "-w",
#     "--watch",
#     default=False,
#     is_flag=True,
#     show_default=True,
#     help="watches a folder for changes",
# )
# @click.argument(
#     "input",
#     type=click.Path(
#         exists=True,
#         path_type=pathlib.Path,
#         file_okay=False,
#         dir_okay=True,
#         readable=True,
#     ),
# )
# @click.argument(
#     "output",
#     type=click.Path(
#         exists=False,
#         path_type=pathlib.Path,
#         file_okay=False,
#         dir_okay=True,
#         writable=True,
#     ),
# )
# def p(
#     model: str, input: pathlib.Path, output: pathlib.Path, watch: bool, **kwargs
# ) -> None:
#     session = new_session(model)
#
#     def process(each_input: pathlib.Path) -> None:
#         try:
#             mimetype = filetype.guess(each_input)
#             if mimetype is None:
#                 return
#             if mimetype.mime.find("image") < 0:
#                 return
#
#             each_output = (output / each_input.name).with_suffix(".png")
#             each_output.parents[0].mkdir(parents=True, exist_ok=True)
#
#             if not each_output.exists():
#                 each_output.write_bytes(
#                     cast(
#                         bytes,
#                         remove(each_input.read_bytes(), session=session, **kwargs),
#                     )
#                 )
#
#                 if watch:
#                     print(
#                         f"processed: {each_input.absolute()} -> {each_output.absolute()}"
#                     )
#         except Exception as e:
#             print(e)
#
#     inputs = list(input.glob("**/*"))
#     if not watch:
#         inputs = tqdm(inputs)
#
#     for each_input in inputs:
#         if not each_input.is_dir():
#             process(each_input)
#
#     if watch:
#         observer = Observer()
#
#         class EventHandler(FileSystemEventHandler):
#             def on_any_event(self, event: FileSystemEvent) -> None:
#                 if not (
#                     event.is_directory or event.event_type in ["deleted", "closed"]
#                 ):
#                     process(pathlib.Path(event.src_path))
#
#         event_handler = EventHandler()
#         observer.schedule(event_handler, input, recursive=False)
#         observer.start()
#
#         try:
#             while True:
#                 time.sleep(1)
#
#         finally:
#             observer.stop()
#             observer.join()
#
#
# @main.command(help="for a http server")
# @click.option(
#     "-p",
#     "--port",
#     default=5000,
#     type=int,
#     show_default=True,
#     help="port",
# )
# @click.option(
#     "-l",
#     "--log_level",
#     default="info",
#     type=str,
#     show_default=True,
#     help="log level",
# )
# @click.option(
#     "-t",
#     "--threads",
#     default=None,
#     type=int,
#     show_default=True,
#     help="number of worker threads",
# )
# def s(port: int, log_level: str, threads: int) -> None:
#     sessions: dict[str, BaseSession] = {}
#     tags_metadata = [
#         {
#             "name": "Background Removal",
#             "description": "Endpoints that perform background removal with different image sources.",
#             "externalDocs": {
#                 "description": "GitHub Source",
#                 "url": "https://github.com/danielgatis/rembg",
#             },
#         },
#     ]
#     app = FastAPI(
#         title="Rembg",
#         description="Rembg is a tool to remove images background. That is it.",
#         version=_version.get_versions()["version"],
#         contact={
#             "name": "Daniel Gatis",
#             "url": "https://github.com/danielgatis",
#             "email": "danielgatis@gmail.com",
#         },
#         license_info={
#             "name": "MIT License",
#             "url": "https://github.com/danielgatis/rembg/blob/main/LICENSE.txt",
#         },
#         openapi_tags=tags_metadata,
#     )
#
#     app.add_middleware(
#         CORSMiddleware,
#         allow_credentials=True,
#         allow_origins=["*"],
#         allow_methods=["*"],
#         allow_headers=["*"],
#     )
#
#     class ModelType(str, Enum):
#         u2net = "u2net"
#         u2netp = "u2netp"
#         u2net_human_seg = "u2net_human_seg"
#         u2net_cloth_seg = "u2net_cloth_seg"
#         silueta = "silueta"
#
#     class CommonQueryParams:
#         def __init__(
#             self,
#             model: ModelType = Query(
#                 default=ModelType.u2net,
#                 description="Model to use when processing image",
#             ),
#             a: bool = Query(default=False, description="Enable Alpha Matting"),
#             af: int = Query(
#                 default=240,
#                 ge=0,
#                 le=255,
#                 description="Alpha Matting (Foreground Threshold)",
#             ),
#             ab: int = Query(
#                 default=10,
#                 ge=0,
#                 le=255,
#                 description="Alpha Matting (Background Threshold)",
#             ),
#             ae: int = Query(
#                 default=10, ge=0, description="Alpha Matting (Erode Structure Size)"
#             ),
#             om: bool = Query(default=False, description="Only Mask"),
#             ppm: bool = Query(default=False, description="Post Process Mask"),
#         ):
#             self.model = model
#             self.a = a
#             self.af = af
#             self.ab = ab
#             self.ae = ae
#             self.om = om
#             self.ppm = ppm
#
#     class CommonQueryPostParams:
#         def __init__(
#             self,
#             model: ModelType = Form(
#                 default=ModelType.u2net,
#                 description="Model to use when processing image",
#             ),
#             a: bool = Form(default=False, description="Enable Alpha Matting"),
#             af: int = Form(
#                 default=240,
#                 ge=0,
#                 le=255,
#                 description="Alpha Matting (Foreground Threshold)",
#             ),
#             ab: int = Form(
#                 default=10,
#                 ge=0,
#                 le=255,
#                 description="Alpha Matting (Background Threshold)",
#             ),
#             ae: int = Form(
#                 default=10, ge=0, description="Alpha Matting (Erode Structure Size)"
#             ),
#             om: bool = Form(default=False, description="Only Mask"),
#             ppm: bool = Form(default=False, description="Post Process Mask"),
#         ):
#             self.model = model
#             self.a = a
#             self.af = af
#             self.ab = ab
#             self.ae = ae
#             self.om = om
#             self.ppm = ppm
#
#     def im_without_bg(content: bytes, commons: CommonQueryParams) -> Response:
#         return Response(
#             remove(
#                 content,
#                 session=sessions.setdefault(
#                     commons.model.value, new_session(commons.model.value)
#                 ),
#                 alpha_matting=commons.a,
#                 alpha_matting_foreground_threshold=commons.af,
#                 alpha_matting_background_threshold=commons.ab,
#                 alpha_matting_erode_size=commons.ae,
#                 only_mask=commons.om,
#                 post_process_mask=commons.ppm,
#             ),
#             media_type="image/png",
#         )
#
#     @app.on_event("startup")
#     def startup():
#         if threads is not None:
#             from anyio import CapacityLimiter
#             from anyio.lowlevel import RunVar
#
#             RunVar("_default_thread_limiter").set(CapacityLimiter(threads))
#
#     @app.get(
#         path="/",
#         tags=["Background Removal"],
#         summary="Remove from URL",
#         description="Removes the background from an image obtained by retrieving an URL.",
#     )
#     async def get_index(
#         url: str = Query(
#             default=..., description="URL of the image that has to be processed."
#         ),
#         commons: CommonQueryParams = Depends(),
#     ):
#         async with aiohttp.ClientSession() as session:
#             async with session.get(url) as response:
#                 file = await response.read()
#                 return await asyncify(im_without_bg)(file, commons)
#
#     @app.post(
#         path="/",
#         tags=["Background Removal"],
#         summary="Remove from Stream",
#         description="Removes the background from an image sent within the request itself.",
#     )
#     async def post_index(
#         file: bytes = File(
#             default=...,
#             description="Image file (byte stream) that has to be processed.",
#         ),
#         commons: CommonQueryPostParams = Depends(),
#     ):
#         return await asyncify(im_without_bg)(file, commons)
#
#     uvicorn.run(app, host="0.0.0.0", port=port, log_level=log_level)



# from typing import List
#
# import numpy as np
# from PIL import Image
# from PIL.Image import Image as PILImage
# from scipy.special import log_softmax
#
# from .session_base import BaseSession
#
# pallete1 = [
#     0,
#     0,
#     0,
#     255,
#     255,
#     255,
#     0,
#     0,
#     0,
#     0,
#     0,
#     0,
# ]
#
# pallete2 = [
#     0,
#     0,
#     0,
#     0,
#     0,
#     0,
#     255,
#     255,
#     255,
#     0,
#     0,
#     0,
# ]
#
# pallete3 = [
#     0,
#     0,
#     0,
#     0,
#     0,
#     0,
#     0,
#     0,
#     0,
#     255,
#     255,
#     255,
# ]
#
#
# class ClothSession(BaseSession):
#     def predict(self, img: PILImage) -> List[PILImage]:
#         ort_outs = self.inner_session.run(
#             None, self.normalize(img, (0.5, 0.5, 0.5), (0.5, 0.5, 0.5), (768, 768))
#         )
#
#         pred = ort_outs
#         pred = log_softmax(pred[0], 1)
#         pred = np.argmax(pred, axis=1, keepdims=True)
#         pred = np.squeeze(pred, 0)
#         pred = np.squeeze(pred, 0)
#
#         mask = Image.fromarray(pred.astype("uint8"), mode="L")
#         mask = mask.resize(img.size, Image.LANCZOS)
#
#         masks = []
#
#         mask1 = mask.copy()
#         mask1.putpalette(pallete1)
#         mask1 = mask1.convert("RGB").convert("L")
#         masks.append(mask1)
#
#         mask2 = mask.copy()
#         mask2.putpalette(pallete2)
#         mask2 = mask2.convert("RGB").convert("L")
#         masks.append(mask2)
#
#         mask3 = mask.copy()
#         mask3.putpalette(pallete3)
#         mask3 = mask3.convert("RGB").convert("L")
#         masks.append(mask3)
#
#         return masks
